import warnings

import numpy as np
import scipy
import scipy.optimize
import scipy.sparse
import scipy.signal
import sklearn
import matplotlib.pyplot as plt
import torch
from tqdm import tqdm

from .. import helpers, util

class Clusterer(util.ROICaT_Module):
    """
    Class for clustering algorithms.
    Performs:
        - Optimal mixing and pruning of similarity matrices:
            - self.find_optimal_parameters_for_pruning()
            - self.make_pruned_similarity_graphs()
        - Clustering:
            - self.fit(): Which uses a modified HDBSCAN
            - self.fit_sequentialHungarian: Which uses a method
             similar to CaImAn's clustering method.
        - Quality control:
            - self.compute_cluster_quality_metrics()

    Initialization ingests and stores similarity matrices.

    RH 2022-2023

    Args:
        s_sf (scipy.sparse.csr_matrix):
            Similarity matrix for spatial footprints.
            Shape: (n_rois, n_rois). Symmetric.
            Expecting input to be manhattan distance of
                spatial footprints normalized between 0 and 1.
        s_NN_z (scipy.sparse.csr_matrix):
            Z-scored similarity matrix for neural network
                output similaries.
            Shape: (n_rois, n_rois). Non-symmetric.
            Expecting input to be the cosine similarity 
                matrix, z-scored row-wise.
        s_SWT_z (scipy.sparse.csr_matrix):
            Z-scored similarity matrix for scattering 
                transform output similarities.
            Shape: (n_rois, n_rois). Non-symmetric.
            Expecting input to be the cosine similarity
                matrix, z-scored row-wise.
        s_sesh (scipy.sparse.csr_matrix, boolean):
            Similarity matrix for session similarity.
            Shape: (n_rois, n_rois). Symmetric.
            Expecting input to be boolean, with 1s where
                the two ROIs are from DIFFERENT sessions.
        verbose (bool):
            Whether to print out information about the
                clustering process.
    """
    def __init__(
        self,
        s_sf=None,
        s_NN_z=None,
        s_SWT_z=None,
        s_sesh=None,
        verbose=True,
    ):
        """
        """
        ## Imports
        super().__init__()

        self.s_sf = s_sf
        self.s_NN_z = s_NN_z
        self.s_SWT_z = s_SWT_z
        self.s_sesh = s_sesh

        ## assert that all feature similarity matrices have the same nnz
        assert self.s_sf.nnz == self.s_NN_z.nnz == self.s_SWT_z.nnz

        self.s_sesh_inv = (self.s_sf != 0).astype(np.bool_)
        self.s_sesh_inv[self.s_sesh.astype(np.bool_)] = False
        self.s_sesh_inv.eliminate_zeros()

        self.s_sesh = self.s_sesh.tolil()
        self.s_sesh[range(self.s_sesh.shape[0]), range(self.s_sesh.shape[1])] = 0
        self.s_sesh = self.s_sesh.tocsr()

        self._verbose = verbose

    def find_optimal_parameters_for_pruning(
        self,
        n_bins=50,
        smoothing_window_bins=5,
        find_parameters_automatically=True,
        kwargs_findParameters={
            'n_patience': 100,
            'tol_frac': 0.05,
            'max_trials': 350,
            'max_duration': 60*10,
            'verbose': False,
        },
        bounds_findParameters={
            'power_SF': (0.3, 2),
            'power_NN': (0.2, 2),
            'power_SWT': (0.1, 1),
            'p_norm': (-5, 5),
            'sig_NN_kwargs_mu': (0, 0.5),
            'sig_NN_kwargs_b': (0.05, 2),
            'sig_SWT_kwargs_mu': (0, 0.5),
            'sig_SWT_kwargs_b': (0.05, 2),
        },
        n_jobs_findParameters=-1,
    ):
        """
        Find the optimal parameters for pruning the similarity graph.
        How this function works:
            1. Make a conjunctive distance matrix using a set of parameters
             for the self.make_conjunctive_distance_matrix function.
            2. Estimates the distribution of pairwise distances
             between ROIs assumed to be the same and those assumed to be
             different ROIs. This is done by comparing the difference in the 
             distribution of pairwise distances between ROIs from the same
             session and those from different sessions. Ideally, the
             main difference will be the presence of 'same' ROIs in the
             inter-session distribution. 
            3. The optimal parameters are then updated using optuna in order
             to maximize the separation between the 'same' and 'different'
             distributions.

        Args:
            n_bins (int):
                Number of bins to use when estimating the distributions. Using
                 a large number of bins makes finding the separation point more
                 noisy, and only slightly more accurate.
            smoothing_window_bins (int):
                Number of bins to use when smoothing the distributions. Using
                 a small number of bins makes finding the separation point more
                 noisy, and only slightly more accurate. Aim for 5-10% of the
                 number of bins.
            kwargs_findParameters (dict):
                Keyword arguments for the Convergence_checker class __init__.
            bounds_findParameters (dict):
                Bounds for the parameters to be optimized.
            n_jobs_findParameters (int):
                Number of jobs to use when finding the optimal parameters.
                If -1, use all available cores.

        Returns:
            kwargs_makeConjunctiveDistanceMatrix_best (dict):
                The optimal (for pruning) keyword arguments using the
                 self.make_conjunctive_distance_matrix function.
        """
        import optuna
        self.bounds_findParameters = bounds_findParameters

        self.n_bins = max(min(self.s_sf.nnz // 30000, 1000), 30) if n_bins is None else n_bins
        self.smooth_window = self.n_bins // 10 if smoothing_window_bins is None else smoothing_window_bins
        # print(f'Pruning similarity graphs with {self.n_bins} bins and smoothing window {smoothing_window}...') if self._verbose else None

        print('Finding mixing parameters using automated hyperparameter tuning...') if self._verbose else None
        optuna.logging.set_verbosity(optuna.logging.WARNING)
        self.checker = Convergence_checker(**kwargs_findParameters)
        self.study = optuna.create_study(direction='minimize', sampler=optuna.samplers.TPESampler(n_startup_trials=kwargs_findParameters['n_patience']//2))
        self.study.optimize(self._objectiveFn_distSameMagnitude, n_jobs=n_jobs_findParameters, callbacks=[self.checker.check])

        self.best_params = self.study.best_params.copy()
        [self.best_params.pop(p) for p in [
            # 'sig_SF_kwargs_mu',
            # 'sig_SF_kwargs_b',
            'sig_NN_kwargs_mu',
            'sig_NN_kwargs_b',
            'sig_SWT_kwargs_mu',
            'sig_SWT_kwargs_b'] if p in self.best_params.keys()]
        # # self.best_params['sig_SF_kwargs'] = {'mu': self.study.best_params['sig_SF_kwargs_mu'],
        # #                                 'b': self.study.best_params['sig_SF_kwargs_b'],}
        # self.best_params['sig_SF_kwargs'] = None
        self.best_params['sig_NN_kwargs'] = {'mu': self.study.best_params['sig_NN_kwargs_mu'],
                                        'b': self.study.best_params['sig_NN_kwargs_b'],}
        self.best_params['sig_SWT_kwargs'] = {'mu': self.study.best_params['sig_SWT_kwargs_mu'],
                                            'b': self.study.best_params['sig_SWT_kwargs_b'],}

        self.kwargs_makeConjunctiveDistanceMatrix_best={
            'power_SF': None,
            'power_NN': None,
            'power_SWT': None,
            'p_norm': None,
            'sig_SF_kwargs': None,
            'sig_NN_kwargs': None,
            'sig_SWT_kwargs': None,
        }
        self.kwargs_makeConjunctiveDistanceMatrix_best.update(self.best_params)
        print(f'Best value found: {self.study.best_value} with parameters {self.best_params}') if self._verbose else None
        return self.kwargs_makeConjunctiveDistanceMatrix_best

    def make_pruned_similarity_graphs(
        self,
        convert_to_probability=False,
        stringency=1.0,
        kwargs_makeConjunctiveDistanceMatrix=None,
        d_cutoff=None,
    ):
        """
        Make pruned similarity graphs.

        Args:
            convert_to_probability (bool):
                Whether to convert the distance and similarity graphs
                 to probability, p(different) and p(same), respectively.
            stringency (float):
                This value changes the threshold for pruning the distance
                 matrix. A higher value will result in less pruning, and a
                 lower value will result in more pruning. The value will be
                 multiplied by the inferred threshold to get the new one.
            kwargs_makeConjunctiveDistanceMatrix (dict):
                Keyword arguments for the self.make_conjunctive_distance_matrix
                 function. If None, the best parameters found using
                 self.find_optimal_parameters will be used.
            d_cutoff (float):
                The cutoff distance for pruning the distance matrix. If None,
                 then the optimal cutoff distance will be inferred.
        """
        if kwargs_makeConjunctiveDistanceMatrix is None:
            if hasattr(self, 'kwargs_makeConjunctiveDistanceMatrix_best'):
                kwargs_makeConjunctiveDistanceMatrix = self.kwargs_makeConjunctiveDistanceMatrix_best
            else:
                kwargs_makeConjunctiveDistanceMatrix = {
                    'power_SF': 0.5,
                    'power_NN': 1.0,
                    'power_SWT': 0.1,
                    'p_norm': -4.0,
                    'sig_SF_kwargs': {'mu':0.5, 'b':0.5},
                    'sig_NN_kwargs': {'mu':0.5, 'b':0.5},
                    'sig_SWT_kwargs': {'mu':0.5, 'b':0.5},
                }
                warnings.warn(f'No kwargs_makeConjunctiveDistanceMatrix provided. Using default parameters: {kwargs_makeConjunctiveDistanceMatrix}')

        self.dConj, self.sConj, sSF_data, sNN_data, sSWT_data, sConj_data = self.make_conjunctive_distance_matrix(
            s_sf=self.s_sf,
            s_NN=self.s_NN_z,
            s_SWT=self.s_SWT_z,
            s_sesh=None,
            **kwargs_makeConjunctiveDistanceMatrix
        )
        dens_same_crop, dens_same, dens_diff, dens_all, edges, d_crossover = self._separate_diffSame_distributions(self.dConj)

        if convert_to_probability:        
            ## convert into probabilities
            ### first smooth dens_diff. (dens_same is already smoothed)
            dens_diff_smooth = self._fn_smooth(dens_diff)
            ### second, compute the probability of each bin
            prob_same = (dens_same / (dens_same + dens_diff_smooth)).numpy()
            ### force to be monotonic decreasing
            prob_same = np.maximum.accumulate(prob_same[::-1])[::-1]
            ### third, append 0 to the end
            prob_same = np.append(prob_same, 0)
            ### third, convert self.dConj to probabilities using interpolation
            import scipy.interpolate
            fn_interp = scipy.interpolate.interp1d(edges, prob_same, kind='linear', fill_value='extrapolate')
            self.sConj.data = fn_interp(self.dConj.data)
            self.dConj.data = 1 - self.sConj.data
            d_crossover = 1 - fn_interp(d_crossover)

        self.distributions_mixing = {
            'kwargs_makeConjunctiveDistanceMatrix': kwargs_makeConjunctiveDistanceMatrix,
            'dens_same_crop': dens_same_crop,
            'dens_same': dens_same,
            'dens_diff': dens_diff,
            'dens_all': dens_all,
            'edges': edges,
            'd_crossover': d_crossover,
        }

        ssf, snn, sswt, ssesh = self.s_sf.copy(), self.s_NN_z.copy(), self.s_SWT_z.copy(), self.s_sesh.copy()
        # ssf.data[ssf.data == 0] = 1e-10
        # snn.data[snn.data == 0] = 1e-10
        # sswt.data[sswt.data == 0] = 1e-10

        min_d = np.nanmin(self.dConj.data)
        range_d = d_crossover - min_d
        self.d_cutoff = min_d + range_d * stringency if d_cutoff is None else d_cutoff
        print(f'Pruning similarity graphs with d_cutoff = {self.d_cutoff}...') if self._verbose else None

        self.graph_pruned = self.dConj.copy()
        self.graph_pruned.data = self.graph_pruned.data < self.d_cutoff
        self.graph_pruned.eliminate_zeros()
        
        def prune(s, graph_pruned):
            import scipy.sparse
            if s is None:
                return None
            s_pruned = scipy.sparse.csr_matrix(graph_pruned.shape, dtype=np.float32)
            s_pruned[graph_pruned] = s[graph_pruned]
            s_pruned = s_pruned.tocsr()
            return s_pruned

        self.s_sf_pruned, self.s_NN_pruned, self.s_SWT_pruned, self.s_sesh_pruned = tuple([prune(s, self.graph_pruned) for s in [ssf, snn, sswt, ssesh]])
        self.dConj_pruned, self.sConj_pruned = prune(self.dConj, self.graph_pruned), prune(self.sConj, self.graph_pruned)

    def fit(
        self,
        d_conj,
        session_bool,
        min_cluster_size=2,
        n_iter_violationCorrection=5,
        cluster_selection_method='leaf',
        d_clusterMerge=None,
        alpha=0.999,
        split_intraSession_clusters=True,
        discard_failed_pruning=True,
        n_steps_clusterSplit=100,
    ):
        """
        Fit clustering using a modified HDBSCAN clustering algorithm.
        The idea here is to use HDBSCAN but avoid having clusters with
         multiple ROIs from the same session. This is accomplished by
         repeating three steps:
            1. Fit HDBSCAN to the data
            2. Identify clusters that have multiple ROIs from the same session
             and walk back down the dendrogram until those clusters are split
             up into non-violating clusters.
            3. Disconnect graph edges between ROIs within each new cluster and
             all other ROIs outside the cluster that are from the same session.

        Args:
            d_conj (float):
                Conjunctive distance matrix.
            session_bool (np.ndarray of bool):
                Boolean array indicating which ROIs belong to which session.
                shape: (n_rois, n_sessions)
            min_cluster_size (int):
                Minimum cluster size to be considered a cluster.
            n_iter_violationCorrection (int):
                Number of iterations to correct for clusters with multiple ROIs per session.
                After cluster splitting, this will go through each cluster and disconnect
                 the ROIs in that cluster from all other ROIs that share a session with any
                 ROI in that cluster. Then it will re-run the HDBSCAN clustering algorithm.
                This is done to overcome the issues with single-linkage clustering finding
                 clusters with multiple ROIs per session.
                Usually this converges after about 5 iterations, and n_sessions at most.
            d_clusterMerge (float):
                Distance threshold for merging clusters.
                All clusters with ROIs closer than this distance will be merged.
                If None, the distance is calculated as the mean + 1*std of the 
                 conjunctive distances.
            cluster_selection_method (str):
                Cluster selection method. Either 'leaf' or 'eom'.
                'leaf' tends towards smaller clusters, 'eom' towards larger clusters.
                See HDBSCAN documentation: 
                 https://hdbscan.readthedocs.io/en/latest/parameter_selection.html
            alpha (float):
                Alpha value. Avoid messing with this if possible.
                Smaller values will result in more clusters.
                See HDBSCAN documentation: 
                 https://hdbscan.readthedocs.io/en/latest/parameter_selection.html
            split_intraSession_clusters (bool):
                If True, clusters that contain ROIs from multiple sessions will be split.
                Only set to False if you want clusters containing multiple
                 ROIs from the same session.
            discard_failed_pruning (bool):
                If True, clusters that fail to prune will be set to -1.
            n_steps_clusterSplit (float):
                Number of steps for splitting clusters with multiple ROIs from
                 the same session. Lower values are faster but less accurate.

        Returns:
            labels (np.ndarray of int):
                Cluster labels for each ROI.
                shape: (n_rois_total)
        """
        import hdbscan
        d = d_conj.copy().multiply(self.s_sesh)

        if d.nnz == 0:
            print('No edges in graph. Returning all -1 labels.') if self._verbose else None
            self.labels = np.ones(d.shape[0], dtype=int) * -1
            return self.labels
            

        print('Fitting with HDBSCAN and splitting clusters with multiple ROIs per session') if self._verbose else None
        for ii in tqdm(range(n_iter_violationCorrection)):
            ## Prep parameters for splitting clusters
            d_clusterMerge = float(np.mean(d.data) + 1*np.std(d.data)) if d_clusterMerge is None else float(d_clusterMerge)
            n_steps_clusterSplit = int(n_steps_clusterSplit)

            n_sessions = session_bool.shape[1]
            max_dist=(d.max() - d.min()) * 1000

            self.hdbs = hdbscan.HDBSCAN(
                min_cluster_size=min_cluster_size,
            #     min_samples=None,
                cluster_selection_epsilon=d_clusterMerge,
                max_cluster_size=n_sessions,
                metric='precomputed',
                alpha=alpha,
            #     p=None,
                algorithm='generic',
            #     leaf_size=100,
            # #     memory=Memory(location=None),
            #     approx_min_span_tree=False,
            #     gen_min_span_tree=False,
            #     core_dist_n_jobs=mp.cpu_count(),
                cluster_selection_method=cluster_selection_method,
            #     allow_single_cluster=False,
            #     prediction_data=False,
            #     match_reference_implementation=False,
                max_dist=max_dist
            )

            self.hdbs.fit(attach_fully_connected_node(
                d, 
                dist_fullyConnectedNode=max_dist,
                n_nodes=1,
            ))
            labels = self.hdbs.labels_[:-1]
            self.labels = labels

            print(f'Initial number of violating clusters: {len(np.unique(labels)[np.array([(session_bool[labels==u].sum(0)>1).sum().item() for u in np.unique(labels)]) > 0])}') if self._verbose else None

            ## Split up labels with multiple ROIs per session
            ## The below code is a bit of a mess, but it works.
            ##  It works by iteratively reducing the cutoff distance
            ##  and splitting up violating clusters until there are 
            ##  no more violations.
            if split_intraSession_clusters:
                labels = labels.copy()
                # d_cut = float(d.data.max())

                sb_t = torch.as_tensor(session_bool, dtype=torch.float32) # (n_rois, n_sessions)
                # print(f'num violating clusters: {np.unique(labels)[np.array([(session_bool[labels==u].sum(0)>1).sum().item() for u in np.unique(labels)]) > 0]}')
                n = len(d_conj.data)
                dcd = np.sort(d_conj.data)
                cuts_all = np.sort(np.unique([dcd[0]/2] + [dcd[int(n*ii)] for ii in np.linspace(0., 1., num=n_steps_clusterSplit, endpoint=False)[::-1]] + [dcd[-1]]))[::-1]
                for d_cut in cuts_all:
                    labels_t = torch.as_tensor(labels, dtype=torch.int64)
                    lab_u_t, lab_u_idx_t = torch.unique(labels_t, return_inverse=True) # (n_clusters,), (n_rois,)
                    lab_oneHot_t = helpers.idx_to_oneHot(lab_u_idx_t, dtype=torch.float32)
                    violations_labels = lab_u_t[((sb_t.T @ lab_oneHot_t) > 1.5).sum(0) > 0]
                    violations_labels = violations_labels[violations_labels > -1]

                    if len(violations_labels) == 0:
                        break
                    
                    for l in violations_labels:
                        idx = np.where(labels==l)[0]
                        if d[idx][:,idx].nnz == 0:
                            labels[idx] = -1

                    labels_new = self.hdbs.single_linkage_tree_.get_clusters(
                        cut_distance=d_cut,
                        min_cluster_size=min_cluster_size,
                    )[:-1]
                    
                    idx_toUpdate = np.isin(labels, violations_labels)
                    labels[idx_toUpdate] = labels_new[idx_toUpdate] + labels.max() + 5
                    labels[(labels_new == -1) * idx_toUpdate] = -1
                
                if discard_failed_pruning:
                    # print(f"Setting all clusters with redundant ROIs to label: -1. Number of clusters with redundant ROIs: {len(np.unique(labels[idx_toUpdate]))}") if self._verbose else None
                    # print(f"IDs: {np.unique(labels[idx_toUpdate])}") if self._verbose else None
                    labels[idx_toUpdate] = -1

            l_u = np.unique(labels)
            l_u = l_u[l_u > -1]
            if ii < n_iter_violationCorrection - 1:
                # print(f'Post-separation number of violating clusters: {n_violating_clusters}') if self._verbose else None
                ## find sessions represented in each cluster and set distances to ROIs in those sessions to 1.
                d = d.tocsr()
                for ii, l in enumerate(np.unique(labels)):
                    if l == -1:
                        continue
                    idx = np.where(labels==l)[0]
                    
                    d_sub = d[idx][:,idx]
                    idx_grid = np.meshgrid(idx, idx)
                    ## set distances of ROIs from same session to 0
                    sesh_to_exclude = 1 - (session_bool @ (session_bool[idx].max(0)))  ## make a mask of sessions that are not represented in the cluster
                    d[idx,:] = d[idx,:].multiply(sesh_to_exclude[None,:])  ## set distances to ROIs from sessions represented in the cluster to 1
                    d[:,idx] = d[:,idx].multiply(sesh_to_exclude[:,None])  ## set distances to ROIs from sessions represented in the cluster to 1
                    d[idx_grid[0], idx_grid[1]] = d_sub  ## undo the above for ROIs in the cluster
                d = d.tocsr()
                d.eliminate_zeros()  ## remove zeros


        labels = helpers.squeeze_integers(labels)
        
        violations_labels = np.unique(labels)[np.array([(session_bool[labels==u].sum(0)>1).sum().item() for u in np.unique(labels)]) > 0]
        violations_labels = violations_labels[violations_labels > -1]
        self.violations_labels = violations_labels

        ## Set clusters with too few ROIs to -1
        u, c = np.unique(labels, return_counts=True)
        labels[np.isin(labels, u[c<2])] = -1
        labels = helpers.squeeze_integers(labels)

        self.labels = labels
        return self.labels

    def fit_sequentialHungarian(
        self,
        d_conj,
        session_bool,
        thresh_cost=0.95,
    ):
        """
        Use CaImAn's method for clustering.
        See their paper and repo for details:
            https://elifesciences.org/articles/38173#s4
            https://github.com/flatironinstitute/CaImAn
            https://github.com/flatironinstitute/CaImAn/blob/master/caiman/base/rois.py

        Args:
            d_conj (scipy.sparse.csr_matrix):
                Distance matrix. Shape (n_rois, n_rois)
            session_bool (np.ndarray): 
                Ahape (n_rois, n_sessions) boolean array indicating which ROIs
                 are in which sessions
            thresh_cost (float, optional):
                Threshold below which ROI pairs are considered potential matches.
        """
        print(f"Clustering with CaImAn's sequential Hungarian algorithm method...") if self._verbose else None
        def find_matches(D_s):
            # todo todocument

            matches = []
            costs = []
            for ii, D in enumerate(D_s):
                # we make a copy not to set changes in the original
                DD = D.copy()
                if np.sum(np.where(np.isnan(DD))) > 0:
                    raise Exception('Distance Matrix contains invalid value NaN')

                # we do the hungarian
                indexes = scipy.optimize.linear_sum_assignment(DD)
                indexes2 = [(ind1, ind2) for ind1, ind2 in zip(indexes[0], indexes[1])]
                matches.append(indexes)
                total = []
                # we want to extract those informations from the hungarian algo
                for row, column in indexes2:
                    value = DD[row, column]
                    total.append(value)
                costs.append(total)
                # send back the results in the format we want
            return matches, costs

        n_roi = session_bool.sum(0)
        n_roi_cum = np.concatenate(([0], np.cumsum(n_roi)))
        
        matchings = []
        matchings.append(list(range(n_roi[0])))

        idx_union = np.arange(n_roi[0])

        for i_sesh in tqdm(range(1,len(n_roi))):
            
            idx_sess = np.arange(n_roi_cum[i_sesh], n_roi_cum[i_sesh+1])
            
            d_sub = d_conj[idx_sess][:, idx_union]
            D = np.ones((len(idx_sess), len(idx_union)))*np.logical_not((d_sub != 0).toarray())*1 + d_sub.toarray()
            D = [D]
            
            matches, costs = find_matches(D)
            matches = matches[0]
            costs = costs[0]

            # store indices
            idx_tp = np.where(np.array(costs) < thresh_cost)[0]
            if len(idx_tp) > 0:
                matched_ROIs1 = matches[0][idx_tp]     # ground truth
                matched_ROIs2 = matches[1][idx_tp]     # algorithm - comp
                non_matched1 = np.setdiff1d(list(range(D[0].shape[0])), matches[0][idx_tp])
                non_matched2 = np.setdiff1d(list(range(D[0].shape[1])), matches[1][idx_tp])
                TP = np.sum(np.array(costs) < thresh_cost) * 1.
            else:
                TP = 0.
                matched_ROIs1 = []
                matched_ROIs2 = []
                non_matched1 = list(range(D[0].shape[0]))
                non_matched2 = list(range(D[0].shape[1]))

            # compute precision and recall
            FN = D[0].shape[0] - TP
            FP = D[0].shape[1] - TP
            TN = 0

            performance = dict()
            performance['recall'] = TP / (TP + FN)
            performance['precision'] = TP / (TP + FP)
            performance['accuracy'] = (TP + TN) / (TP + FP + FN + TN)
            performance['f1_score'] = 2 * TP / (2 * TP + FP + FN)

            mat_sess, mat_un, nm_sess, nm_un, performance, A2_len = matched_ROIs1, matched_ROIs2, non_matched1, non_matched2, performance, len(idx_union)
            
            idx_union[mat_un] = idx_sess[mat_sess]
            idx_union = np.concatenate((idx_union, idx_sess[nm_sess]))
            
            new_match = np.zeros(n_roi[i_sesh], dtype=int)
            new_match[mat_sess] = mat_un
            new_match[nm_sess] = range(A2_len, len(idx_union))
            matchings.append(new_match.tolist())

        self.seqHung_performance = performance

        labels = np.concatenate(matchings)
        u, c = np.unique(labels, return_counts=True)
        labels[np.isin(labels, u[c == 1])] = -1
        labels = helpers.squeeze_integers(labels)
        self.labels = labels
        return self.labels
            
    @classmethod
    def make_conjunctive_distance_matrix(
        cls,
        s_sf=None,
        s_NN=None,
        s_SWT=None,
        s_sesh=None,
        power_SF=1,
        power_NN=1,
        power_SWT=1,
        p_norm=1,
        sig_SF_kwargs={'mu':0.5, 'b':0.5},
        sig_NN_kwargs={'mu':0.5, 'b':0.5},
        sig_SWT_kwargs={'mu':0.5, 'b':0.5},
    ):
        """
        Make a distance matrix from the three similarity matrices.

        Args:
            s_sf (scipy.sparse.csr_matrix):
                Similarity matrix for spatial footprints.
            s_NN_z (scipy.sparse.csr_matrix):
                Z-scored similarity matrix for neural network features.
            s_SWT_z (scipy.sparse.csr_matrix):
                Z-scored similarity matrix for scattering wavelet 
                 transform features.
            power_SF (float):
                Power to which to raise the spatial footprint similarity.
            power_NN (float):
                Power to which to raise the neural network similarity.
            power_SWT (float):
                Power to which to raise the scattering wavelet transform 
                 similarity.
            p_norm (float):
                p-norm to use for the conjunction of the similarity
                 matrices.
            sig_SF_kwargs (dict):
                Keyword arguments for the sigmoid function applied to the
                 spatial footprint overlap similarity matrix.
                See helpers.generalised_logistic_function for details.
            sig_NN_kwargs (dict):
                Keyword arguments for the sigmoid function applied to the
                 neural network similarity matrix.
                See helpers.generalised_logistic_function for details.
            sig_SWT_kwargs (dict):
                Keyword arguments for the sigmoid function applied to the
                 scattering wavelet transform similarity matrix.
                See helpers.generalised_logistic_function for details.
            plot_sigmoid (bool):
                Whether to plot the sigmoid functions applied to the
                 neural network and scattering wavelet transform
                 similarity matrices.
        """
        assert (s_sf is not None) or (s_NN is not None) or (s_SWT is not None), \
            'At least one of s_sf, s_NN, or s_SWT must be provided.'
        
        p_norm = 1e-9 if p_norm == 0 else p_norm

        sSF_data = cls._activation_function(s_sf.data, sig_SF_kwargs, power_SF) if s_sf is not None else None
        sNN_data = cls._activation_function(s_NN.data, sig_NN_kwargs, power_NN) if s_NN is not None else None
        sSWT_data = cls._activation_function(s_SWT.data, sig_SWT_kwargs, power_SWT) if s_SWT is not None else None

        s_list = [s for s in [sSF_data, sNN_data, sSWT_data] if s is not None]
        
        sConj_data = cls._pNorm(
            s_list=s_list,
            p=p_norm,
        )
        # sConj_data = sConj_data * s_sesh.data if s_sesh is not None else sConj_data
        # sConj_data = sConj_data * np.logical_not(s_sesh.data) if s_sesh is not None else sConj_data

        ## make sConj
        sConj = s_sf.copy() if s_sf is not None else s_NN.copy() if s_NN is not None else s_SWT.copy()
        sConj.data = sConj_data.numpy() 
        sConj = sConj.multiply(s_sesh) if s_sesh is not None else sConj
        # sConj.eliminate_zeros()

        ## make dConj
        dConj = sConj.copy()
        dConj.data = 1 - dConj.data

        return dConj, sConj, sSF_data, sNN_data, sSWT_data, sConj_data

    @classmethod
    def _activation_function(cls, s, sig_kwargs={'mu':0.0, 'b':1.0}, power=1):
        if s is None:
            return None
        if (sig_kwargs is not None) and (power is not None):
            return helpers.generalised_logistic_function(torch.as_tensor(s, dtype=torch.float32), **sig_kwargs)**power
        elif (sig_kwargs is None) and (power is not None):
            return torch.maximum(torch.as_tensor(s, dtype=torch.float32), torch.as_tensor([0], dtype=torch.float32))**power
            # return torch.as_tensor(s, dtype=torch.float32)**power
        elif (sig_kwargs is not None) and (power is None):
            return helpers.generalised_logistic_function(torch.as_tensor(s, dtype=torch.float32), **sig_kwargs)
        else:
            return torch.as_tensor(s, dtype=torch.float32)
        
    @classmethod
    def _pNorm(cls, s_list, p):
        """
        Calculate the p-norm of a list of similarity matrices.
        """
        s_list_noNones = [s for s in s_list if s is not None]
        return (torch.mean(torch.stack(s_list_noNones, axis=0)**p, dim=0))**(1/p)
        # return np.linalg.norm(np.stack(s_list_noNones, axis=0), ord=p, axis=0)

    # def plot_sigmoids(self):
    #     fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(16,4))
    #     axs[0].plot(np.linspace(-0,2,1001), self.sig_sf(np.linspace(-5,5,1001)))
    #     axs[0].set_title('sigmoid SF')
    #     axs[1].plot(np.linspace(-1,1,1001), self.sig_NN(np.linspace(-5,5,1001)))
    #     axs[1].set_title('sigmoid NN')
    #     axs[2].plot(np.linspace(-1,1,1001), self.sig_SWT(np.linspace(-5,5,1001)))
    #     axs[2].set_title('sigmoid SWT')


    def plot_similarity_relationships(
        self, 
        plots_to_show=[1,2,3], 
        max_samples=1000000, 
        kwargs_scatter={'s':1, 'alpha':0.1},
        kwargs_makeConjunctiveDistanceMatrix={
            'power_SF': 0.5,
            'power_NN': 1.0,
            'power_SWT': 0.1,
            'p_norm': -4.0,
            'sig_SF_kwargs': {'mu':0.5, 'b':0.5},
            'sig_NN_kwargs': {'mu':0.5, 'b':0.5},
            'sig_SWT_kwargs': {'mu':0.5, 'b':0.5},
        },
    ):
        """
        Plot the similarity relationships between the three similarity
         matrices.

        Args:
            plots_to_show (list):
                Which plots to show.
                1: Spatial footprints vs. neural network features.
                2: Spatial footprints vs. scattering wavelet transform
                 features.
                3: Neural network features vs. scattering wavelet.
            max_samples (int):
                Maximum number of samples to plot.
                Use smaller numbers for faster plotting.
            kwargs_scatter (dict):
                Keyword arguments for the matplotlib.pyplot.scatter plot.
            kwargs_makeConjunctiveDistanceMatrix (dict):
                Keyword arguments for the makeConjunctiveDistanceMatrix
                 method.

        Returns:
            fig (matplotlib.pyplot.figure):
                Figure object.
            axs (matplotlib.pyplot.axes):
                Axes object.
        """
        dConj, sConj, sSF_data, sNN_data, sSWT_data, sConj_data = self.make_conjunctive_distance_matrix(
            s_sf=self.s_sf,
            s_NN=self.s_NN_z,
            s_SWT=self.s_SWT_z,
            s_sesh=None,
            **kwargs_makeConjunctiveDistanceMatrix
        )

        ## subsample similarities for plotting
        idx_rand = np.floor(np.random.rand(min(max_samples, len(dConj.data))) * len(dConj.data)).astype(int)
        ssf_sub = sSF_data[idx_rand]
        snn_sub = sNN_data[idx_rand]
        sswt_sub = sSWT_data[idx_rand] if sSWT_data is not None else None
        d_conj_sub = dConj.data[idx_rand]

        fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(20,4))
        ## set figure title
        fig.suptitle('Similarity relationships', fontsize=16)
        
        ## plot similarity relationships
        if 1 in plots_to_show:
            axs[0].scatter(ssf_sub, snn_sub, c=d_conj_sub, **kwargs_scatter)
            axs[0].set_xlabel('sim Spatial Footprint')
            axs[0].set_ylabel('sim Neural Network')
        if sSWT_data is not None:
            if 2 in plots_to_show:
                axs[1].scatter(ssf_sub, sswt_sub, c=d_conj_sub, **kwargs_scatter)
                axs[1].set_xlabel('sim Spatial Footprint')
                axs[1].set_ylabel('sim Scattering Wavelet Transform')
            if 3 in plots_to_show:
                axs[2].scatter(snn_sub, sswt_sub, c=d_conj_sub, **kwargs_scatter)
                axs[2].set_xlabel('sim Neural Network')
                axs[2].set_ylabel('sim Scattering Wavelet Transform')
        
        return fig, axs

    def plot_distSame(self, kwargs_makeConjunctiveDistanceMatrix=None):
        """
        Plot the estimated distribution of the pairwise similarities
         between matched ROI pairs of ROIs.
        """
        kwargs = kwargs_makeConjunctiveDistanceMatrix if kwargs_makeConjunctiveDistanceMatrix is not None else self.best_params
        dConj, sConj, sSF_data, sNN_data, sSWT_data, sConj_data = self.make_conjunctive_distance_matrix(
            s_sf=self.s_sf,
            s_NN=self.s_NN_z,
            s_SWT=self.s_SWT_z,
            s_sesh=None,
            **kwargs
        )
        dens_same_crop, dens_same, dens_diff, dens_all, edges, d_crossover = self._separate_diffSame_distributions(dConj)
        # centers = (edges[1:] + edges[:-1]) / 2
        if edges is None:
            print('No crossover found, not plotting')
            return None
        
        plt.figure()
        plt.stairs(dens_same, edges, linewidth=5)
        plt.stairs(dens_same_crop, edges, linewidth=3)
        plt.stairs(dens_diff, edges)
        plt.stairs(dens_all, edges)
        # plt.stairs(dens_diff - dens_same, edges)
        # plt.stairs(dens_all - dens_diff, edges)
        # plt.stairs((dens_diff * dens_same)*1000, edges)
        plt.axvline(d_crossover, color='k', linestyle='--')
        plt.ylim([dens_same.max()*-0.5, dens_same.max()*1.5])
        plt.title('Pairwise similarities')
        plt.xlabel('distance or prob(different)')
        plt.ylabel('counts or density')
        plt.legend(['same', 'same (cropped)', 'diff', 'all', 'diff - same', 'all - diff', '(diff * same) * 1000', 'crossover'])

    def _separate_diffSame_distributions(self, d_conj):
        """
        Estimates the distribution of pairwise similarities for
         'same' and 'different' pairs of ROIs. 
        estimate the 'same' distribution as the different between
         all pairwise distances (includes different and same) and
         intra-session distances (known different). same = all - intra
        See _objectiveFn_distSameMagnitude docstring for more details.

        Args:
            d_conj (scipy.sparse.csr_matrix):
                conjunctive distance matrix.

        Returns:
            dens_same_crop (np.ndarray):
                Distribution density of pairwise similarities 
                 that are assumed to be from the same ROI. It is
                 'cropped' because values below crossover point
                 between same and different distributions are set
                 to zero.
            dens_same (np.ndarray):
                Un-cropped version of dens_same_crop.
            dens_diff (np.ndarray):
                Distribution density of pairwise similarities
                that are assumed to be from different ROIs.
            dens_all (np.ndarray):
                Distribution density of all pairwise similarities.
            edges (np.ndarray):
                Edges of bins used to compute densities.
            d_crossover (float):
                Distance at which the same and different distributions
                 crossover.
        """
        # kernel = torch.ones(1,1,self.smooth_window, device=d_conj.device)/self.smooth_window
        # trace_edgeCorrection = torch.conv1d(
        #     torch.ones(1,1, self.n_bins, device=d_conj.device),
        #     torch.ones(1,1,self.smooth_window, device=d_conj.device)/self.smooth_window, 
        #     padding='same'
        # )[0,0,:]
        # fn_smooth = lambda x: torch.conv1d(x[None,None,:], torch.ones(1,1,self.smooth_window, device=x.device)/self.smooth_window, padding='same')[0,0,:] if self.smooth_window > 1 else x
        
        self._fn_smooth = helpers.Convolver_1d(
            kernel=torch.ones(self.smooth_window),
            length_x=self.n_bins,
            pad_mode='same',
            correct_edge_effects=True,
            device='cpu',
        ).convolve

        edges = torch.linspace(0,1, self.n_bins+1, dtype=torch.float32)
        
        d_all = d_conj.copy()
        counts, _ = torch.histogram(torch.as_tensor(d_all.data, dtype=torch.float32), edges)
        # dens_all = fn_smooth(counts / counts.sum())  ## distances of all pairs of ROIs
        # dens_all = counts / counts.sum()  ## distances of all pairs of ROIs
        dens_all = counts  ## distances of all pairs of ROIs
        # dens_all = counts / counts[-1]  ## distances of all pairs of ROIs

        d_intra = d_conj.multiply(self.s_sesh_inv)
        d_intra.eliminate_zeros()
        counts, _ = torch.histogram(torch.as_tensor(d_intra.data, dtype=torch.float32), edges)
        # dens_diff = fn_smooth(counts / counts.sum())  ## distances of known differents
        # dens_diff = counts / counts.sum()  ## distances of known differents
        dens_diff = counts * (len(d_all.data) / len(d_intra.data))
        # dens_diff = counts / counts[-1]  ## distances of known differents

        dens_same = dens_all - dens_diff  ## estimate the 'same' distribution as the different between all distances (includes different and same) and intra-session distances (known different)
        dens_same = torch.maximum(dens_same, torch.as_tensor([0], dtype=torch.float32))
        # dens_same = torch.as_tensor(scipy.signal.savgol_filter(dens_same.cpu().numpy(), self.smooth_window, 3))  ## smooth the 'same' distribution
        dens_same = self._fn_smooth(dens_same)

        dens_deriv = dens_diff - dens_same  ## difference in 'diff' and 'same' distributions
        dens_deriv[dens_diff.argmax():] = 0
        if torch.where(dens_deriv < 0)[0].shape[0] == 0:  ## if no crossover point, return None
            return None, None, None, None, None, None
        idx_crossover = torch.where(dens_deriv < 0)[0][-1] + 1 
        d_crossover = edges[idx_crossover].item()

        dens_same_crop = dens_same.clone()
        dens_same_crop[idx_crossover:] = 0
        return dens_same_crop, dens_same, dens_diff, dens_all, edges, d_crossover
    def _objectiveFn_distSameMagnitude(self, trial):
        """
        Objective function for Optuna hyperparameter optimization.
        Measures and outputs the magnitude of the 'same' distribution.
        The 'same' distribution is the distribution of distances 
         between pairs of ROIs that are estimated to be the same.
        As the parameters for building the conjuctive distance matrix
         are optimized, the 'same' and 'different' distributions should
         separate from each other. The less the two overlap, the larger
         the effective magnitude of the 'same' distribution.
        
        Args:
            trial (optuna.trial.Trial): 
                Optuna trial object.

        Returns:
            loss (float):
                Magnitude of the 'same' distribution.
        """
        # power_SF = trial.suggest_float('power_SF', *self.bounds_findParameters['power_SF'], log=False)
        power_SF = 1
        power_NN = trial.suggest_float('power_NN', *self.bounds_findParameters['power_NN'], log=False)
        power_SWT = trial.suggest_float('power_SWT', *self.bounds_findParameters['power_SWT'], log=False)
        # power_SWT = 0
        p_norm = trial.suggest_float('p_norm', *self.bounds_findParameters['p_norm'], log=False)
        
        # sig_SF_kwargs = {
        #     'mu':trial.suggest_float('sig_SF_kwargs_mu', 0.1, 0.5, log=False),
        #     'b':trial.suggest_float('sig_SF_kwargs_b', 0.1, 2, log=False),
        # }
        sig_SF_kwargs = None
        sig_NN_kwargs = {
            'mu':trial.suggest_float('sig_NN_kwargs_mu', *self.bounds_findParameters['sig_NN_kwargs_mu'], log=False),
            'b':trial.suggest_float('sig_NN_kwargs_b', *self.bounds_findParameters['sig_NN_kwargs_b'], log=False),
        }
        # sig_NN_kwargs = None
        sig_SWT_kwargs = {
            'mu':trial.suggest_float('sig_SWT_kwargs_mu', *self.bounds_findParameters['sig_SWT_kwargs_mu'], log=False),
            'b':trial.suggest_float('sig_SWT_kwargs_b', *self.bounds_findParameters['sig_SWT_kwargs_b'], log=False),
        }
        # sig_SWT_kwargs = None

        dConj, sConj, sSF_data, sNN_data, sSWT_data, sConj_data = self.make_conjunctive_distance_matrix(
            s_sf=self.s_sf,
            s_NN=self.s_NN_z,
            s_SWT=self.s_SWT_z,
            s_sesh=None,
            power_SF=power_SF,
            power_NN=power_NN,
            power_SWT=power_SWT,
            p_norm=p_norm,
        #     sig_sf_kwargs={'mu':1.0, 'b':0.5},
            sig_SF_kwargs=sig_SF_kwargs,
            sig_NN_kwargs=sig_NN_kwargs,
            sig_SWT_kwargs=sig_SWT_kwargs,
        )
        
        dens_same_crop, dens_same, dens_diff, dens_all, edges, d_crossover = self._separate_diffSame_distributions(dConj)
        if dens_same_crop is None:
            return 0

        # loss = dens_same_crop.sum().item()
        loss = (dens_same * dens_diff).sum().item()
        
        return loss  # Output must be a scalar. Used to update the hyperparameters
    

    def compute_quality_metrics(
        self,
        sim_mat=None,
        dist_mat=None,
        labels=None,
    ):
        """
        Compute quality metrics.
        RH 2023

        Args:
            sim_mat (scipy.sparse.csr_matrix):
                Similarity matrix of shape (n_samples, n_samples).
                If None then self.sConj must exist.
            dist_mat (scipy.sparse.csr_matrix):
                Distance matrix of shape (n_samples, n_samples).
                If None then self.dConj must exist.
            labels (np.ndarray):
                Cluster labels of shape (n_samples,).
                If None, then self.labels must exist.
        """
        if sim_mat is None:
            assert hasattr(self, 'sConj'), "self.sConj does not exist. Run self.find_optimal_parameters_for_pruning() first or specify sim_mat."
            sim_mat = self.sConj
        if dist_mat is None:
            assert hasattr(self, 'dConj'), "self.dConj does not exist. Run self.find_optimal_parameters_for_pruning() first or specify dist_mat."
            dist_mat = self.dConj
        if labels is None:
            assert hasattr(self, 'labels'), "self.labels does not exist. Run self.find_optimal_parameters_for_pruning() first or specify labels."
            labels = self.labels
            
        assert scipy.sparse.issparse(sim_mat), "sim_mat must be a scipy.sparse.csr_matrix."
        assert scipy.sparse.issparse(dist_mat), "dist_mat must be a scipy.sparse.csr_matrix."

        cs_intra_means, cs_intra_mins, cs_intra_maxs, cs_sil = cluster_quality_metrics(
            sim=sim_mat,
            labels=labels,
        )

        import sklearn
        # if sklearn.__version__ < '1.3':
        ## Warn that current version is memory intensive and will be improved when sklearn 1.3 is released
        warnings.warn("Current version of silhouette samples calculation is memory intensive and will be improved when sklearn 1.3 is released.")
        import sparse
        d_dense = sparse.COO(dist_mat.copy().tocsr())
        d_dense.fill_value = (d_dense.max() - d_dense.min()) * 1000
        d_dense = d_dense.todense()
        np.fill_diagonal(d_dense, 0)
        rs_sil = sklearn.metrics.silhouette_samples(X=d_dense, labels=labels, metric='precomputed')

        self.quality_metrics = {
            'cluster_intra_means': cs_intra_means,
            'cluster_intra_mins': cs_intra_mins,
            'cluster_intra_maxs': cs_intra_maxs,
            'cluster_silhouette': cs_sil,
            'sample_silhouette': rs_sil,
            'hdbscan': {
                'sample_outlierScores': self.hdbs.outlier_scores_[:-1],  ## Remove last element which is the outlier score for the new fully connected node
                'sample_probabilities': self.hdbs.probabilities_[:-1],  ## Remove last element which is the outlier score for the new fully connected node
            }  if hasattr(self, 'hdbs') else None,
            'sequentialHungarian': {
                'recall': self.seqHung_performance['recall'],
                'precision': self.seqHung_performance['precision'],
                'f1_score': self.seqHung_performance['f1_score'],
                'accuracy': self.seqHung_performance['accuracy'],
            } if hasattr(self, 'seqHung_performance') else None,
        }
        return self.quality_metrics
        

def attach_fully_connected_node(d, dist_fullyConnectedNode=None, n_nodes=1):
    """
    This function takes in a sparse distance graph (csr_matrix) that has
     more than one component (multiple unconnected subgraphs) and appends
     a single node to the graph that is weakly connected to all nodes.
     
    Args:
        d (scipy.sparse.csr_matrix):
            Sparse graph with multiple components.
            See scipy.sparse.csgraph.connected_components
        dist_fullyConnectedNode (float):
            Value to use for the connection strengh to all other nodes.
            Value will be appended as elements in a new row and column at
             the ends of the 'd' matrix.
            If None, then the value will be set to 1000 times the difference
             between the maximum and minimum values in 'd'.
        n_nodes (int):
            Number of nodes to append to the graph.
             
     Returns:
         d2 (scipy.sparse.csr_matrix):
             Sparse graph with only one component.cluster_quality_metrics
    """
    if dist_fullyConnectedNode is None:
        dist_fullyConnectedNode = (d.max() - d.min()) * 1000
    
    d2 = d.copy()
    d2 = scipy.sparse.vstack((d2, np.ones((n_nodes,d2.shape[1]), dtype=d.dtype)*dist_fullyConnectedNode))
    d2 = scipy.sparse.hstack((d2, np.ones((d2.shape[0],n_nodes), dtype=d.dtype)*dist_fullyConnectedNode))

    return d2.tocsr()

class Convergence_checker:
    """
    A class that is used to check if the optuna optimization has converged.
    """
    def __init__(
        self, 
        n_patience=10, 
        tol_frac=0.05, 
        max_trials=350, 
        max_duration=60*10, 
        verbose=True
    ):
        """
        Args:
            n_patience (int):
                Number of trials to look back to check for convergence.
                Also the minimum number of trials that must be completed
                 before starting to check for convergence.
            tol_frac (float):
                Fractional tolerance for convergence.
                The best output value must change by less than this 
                 fractional amount to be considered converged.
            max_trials (int):
                Maximum number of trials to run before stopping.
            max_duration (float):
                Maximum number of seconds to run before stopping.
            verbose (bool):
                If True, print messages.
        """
        self.bests = []
        self.best = np.inf
        self.n_patience = n_patience
        self.tol_frac = tol_frac
        self.max_trials = max_trials
        self.max_duration = max_duration
        self.num_trial = 0
        self.verbose = verbose
        
    def check(self, study, trial):
        """
        Check if the optuna optimization has converged.
        This function should be used as the callback function for the
         optuna study.

        Args:
            study (optuna.study.Study):
                Optuna study object.
            trial (optuna.trial.FrozenTrial):
                Optuna trial object.
        """
        dur_first, dur_last = study.trials[0].datetime_complete, trial.datetime_complete
        if (dur_first is not None) and (dur_last is not None):
            duration = (dur_last - dur_first).total_seconds()
        else:
            duration = 0
        
        if trial.value < self.best:
            self.best = trial.value
        self.bests.append(self.best)
            
        bests_recent = np.unique(self.bests[-self.n_patience:])
        if self.num_trial > self.n_patience and ((np.abs(bests_recent.max() - bests_recent.min())/np.abs(self.best)) < self.tol_frac):
            print(f'Stopping. Convergence reached. Best value ({self.best*10000}) over last ({self.n_patience}) trials fractionally changed less than ({self.tol_frac})') if self.verbose else None
            study.stop()
        if self.num_trial >= self.max_trials:
            print(f'Stopping. Trial number limit reached. num_trial={self.num_trial}, max_trials={self.max_trials}.') if self.verbose else None
            study.stop()
        if duration > self.max_duration:
            print(f'Stopping. Duration limit reached. study.duration={duration}, max_duration={self.max_duration}.') if self.verbose else None
            study.stop()
            
        if self.verbose:
            print(f'Trial num: {self.num_trial}. Duration: {duration:.3f}s. Best value: {self.best:3e}. Current value:{trial.value:3e}') if self.verbose else None
        self.num_trial += 1


def score_labels(labels_test, labels_true, ignore_negOne=False, thresh_perfect=0.9999999999, compute_mutual_info=False):
    """
    Compute the score of the clustering.
    The best match is found by solving the linear sum assignment problem.
    The score is bounded between 0 and 1.

    Note: The score is not symmetric if the number of true and test
     labels are not the same. I.e. switching labels_test and labels_true
     can lead to different scores. This is not a bug. This is because
     we are scoring how well each true set is matched by an optimally
     assigned test set.

    RH 2022

    Args:
        labels_test (np.array): 
            Labels of the test clusters/sets.
        labels_true (np.array):
            Labels of the true clusters/sets.
        thresh_perfect (float):
            threshold for perfect match.
            Mostly used for numerical stability.

    Returns:
        score_unweighted_partial (float):
            Average correlation between the 
             best matched sets of true and test labels.
        score_unweighted_perfect (float):
            Fraction of perfect matches.
        score_weighted_partial (float):
            Average correlation between the best matched sets of
             true and test labels. Weighted by the number of elements
             in each true set.
        score_weighted_perfect (float):
            Fraction of perfect matches. Weighted by the number of
             elements in each true set.
        hi (np.array):
            'Hungarian Indices'. Indices of the best matched sets.
    """
    assert len(labels_test) == len(labels_true), 'RH ERROR: labels_test and labels_true must be the same length.'
    if ignore_negOne:
        labels_test = np.array(labels_test.copy(), dtype=int)
        labels_true = np.array(labels_true.copy(), dtype=int)
        
        labels_test = labels_test[labels_true > -1].copy()
        labels_true = labels_true[labels_true > -1].copy()

    ## convert labels to boolean
    bool_test = np.stack([labels_test==l for l in np.unique(labels_test)], axis=0).astype(np.float32)
    bool_true = np.stack([labels_true==l for l in np.unique(labels_true)], axis=0).astype(np.float32)

    if bool_test.shape[0] < bool_true.shape[0]:
        bool_test = np.concatenate((bool_test, np.zeros((bool_true.shape[0] - bool_test.shape[0], bool_true.shape[1]))))

    ## compute cross-correlation matrix, and crop to 
    na = bool_true.shape[0]
    cc = np.corrcoef(x=bool_true, y=bool_test)[:na][:,na:]  ## corrcoef returns the concatenated cross-corr matrix (self corr mat along diagonal). The indexing at the end is to extract just the cross-corr mat
    cc[np.isnan(cc)] = 0

    ## find hungarian assignment matching indices
    hi = scipy.optimize.linear_sum_assignment(
        cost_matrix=cc,
        maximize=True,
    )

    ## extract correlation scores of matches
    cc_matched = cc[hi[0], hi[1]]

    ## compute score
    score_weighted_partial = np.sum(cc_matched * bool_true.sum(axis=1)[hi[0]]) / bool_true[hi[0]].sum()
    score_unweighted_partial = np.mean(cc_matched)
    ## compute perfect score
    score_weighted_perfect = np.sum(bool_true.sum(axis=1)[hi[0]] * (cc_matched > thresh_perfect)) / bool_true[hi[0]].sum()
    score_unweighted_perfect = np.mean(cc_matched > thresh_perfect)
    
    ## compute adjusted rand score
    score_rand = sklearn.metrics.adjusted_rand_score(labels_true, labels_test)
    
    ## compute adjusted mutual info score
    score_mutual_info = sklearn.metrics.adjusted_mutual_info_score(labels_true, labels_test) if compute_mutual_info else None

    out = {
        'score_weighted_partial': score_weighted_partial,
        'score_weighted_perfect': score_weighted_perfect,
        'score_unweighted_partial': score_unweighted_partial,
        'score_unweighted_perfect': score_unweighted_perfect,
        'adj_rand_score': score_rand,
        'adj_mutual_info_score': score_mutual_info,
        'ignore_negOne': ignore_negOne,
        'idx_hungarian': hi,
    }
    return out


def cluster_quality_metrics(sim, labels):
    """
    Computes the cluster quality metrics for a clustering solution.
    Computes the intra_means, intra_mins, intra_maxs, and cluster silhouette score
    RH 2023

    Args:
        sim (np.array or scipy.sparse.csr_matrix):
            Similarity matrix of shape (n_roi, n_roi).
            Can be obtained using _, sConj, _,_,_,_ = clusterer.make_conjunctive_similarity_matrix().
        labels (np.array):
            Cluster labels of shape (n_roi,).

    Returns:
        cs_intra_means (np.array):
            Intra-cluster mean similarity of shape (n_clusters,).
        cs_intra_mins (np.array):
            Intra-cluster min similarity of shape (n_clusters,).
        cs_intra_maxs (np.array):
            Intra-cluster max similarity of shape (n_clusters,).
        cs_sil (np.array):
            Cluster silhouette score of shape (n_clusters,).
    """
    import sparse
    
    cs_mean, cs_max, cs_min = helpers.compute_cluster_similarity_matrices(sim, labels, verbose=True)
    fn_sil_score = lambda intra, inter: (intra - inter) / np.maximum(intra, inter)

    eye_inv = 1 - sparse.eye(cs_max.shape[0])

    cs_intra_means = cs_mean.diagonal()
    cs_inter_maxOfMaxs = (eye_inv * cs_max).max(0)
    cs_sil = fn_sil_score(cs_intra_means, cs_inter_maxOfMaxs)
    cs_intra_mins = cs_min.diagonal()
    cs_intra_maxs = cs_max.diagonal()
    
    return cs_intra_means, cs_intra_mins, cs_intra_maxs, cs_sil

def make_label_variants(labels, n_roi_bySession):
    """
    Makes convenient variants of labels.
    RH 2023

    Args:
        labels (np.array):
            Cluster integer labels of shape (n_roi,).
        n_roi_bySession (np.array):
            Number of ROIs in each session.

    Returns:
        labels_squeezed (np.array):
    """
    import scipy.sparse

    ## assert that labels is a 1D np.array or list of numbers
    if isinstance(labels, list):
        labels = np.array(labels, dtype=np.int64)
    elif isinstance(labels, np.ndarray):
        labels = labels.astype(np.int64)
    else:
        raise TypeError('RH ERROR: labels must be a 1D np.array or list of numbers.')
    assert labels.ndim == 1, 'RH ERROR: labels must be a 1D np.array or list of numbers.'

    ## assert that n_roi_bySession is a 1D np.array or list of numbers
    if isinstance(n_roi_bySession, list):
        n_roi_bySession = np.array(n_roi_bySession, dtype=np.int64)
    elif isinstance(n_roi_bySession, np.ndarray):
        n_roi_bySession = n_roi_bySession.astype(np.int64)
    else:
        raise TypeError('RH ERROR: n_roi_bySession must be a 1D np.array or list of numbers.')
    assert n_roi_bySession.ndim == 1, 'RH ERROR: n_roi_bySession must be a 1D np.array or list of numbers.'
    
    ## assert that the number of labels adds up to the number of ROIs
    n_roi_total = n_roi_bySession.sum()
    n_roi_cumsum = np.concatenate([[0], n_roi_bySession.cumsum()])
    assert labels.shape[0] == n_roi_bySession.sum(), 'RH ERROR: the number of labels must add up to the number of ROIs.'

    ## make session_bool
    session_bool = util.make_session_bool(n_roi_bySession)

    ## make variants
    labels_squeezed = helpers.squeeze_integers(labels)
    labels_bySession = [labels_squeezed[idx] for idx in session_bool.T]
    labels_bool = scipy.sparse.vstack([scipy.sparse.csr_matrix(labels_squeezed==u) for u in np.sort(np.unique(labels_squeezed))]).T.tocsr()
    labels_bool_bySession = [labels_bool[idx] for idx in session_bool.T]
    labels_dict = {u: np.where(labels_squeezed==u)[0] for u in np.unique(labels_squeezed)}

    ## testing
    assert np.allclose(np.concatenate(labels_bySession), labels_squeezed)
    assert np.allclose(labels_bool.nonzero()[1] - 1, labels_squeezed)
    assert np.alltrue([np.allclose(np.where(labels_squeezed==u)[0], ldu) for u, ldu in labels_dict.items()])

    return labels_squeezed, labels_bySession, labels_bool, labels_bool_bySession, labels_dict


import copy
import typing
import warnings

import numpy as np
import cv2
import scipy.sparse
from tqdm import tqdm
import matplotlib.pyplot as plt

from .. import helpers, util

class Aligner(util.ROICaT_Module):
    """
    A class for registering ROIs to a template FOV.
    Currently relies on available OpenCV methods for 
     non-rigid registration.
    RH 2023

    Args:
    verbose (bool):
        Whether to print progress updates.
    """
    def __init__(
        self,
        verbose=True,
    ):
        self._verbose = verbose
        
        self.remapingIdx_geo = None
        self.warp_matrices = None

        self.remappingIdx_nonrigid = None

        self._HW = None

    @classmethod
    def augment_FOV_images(
        self,
        ims: typing.List[np.ndarray],
        spatialFootprints: typing.List[scipy.sparse.csr_matrix]=None,
        roi_FOV_mixing_factor: float=0.5,
        use_CLAHE: bool=True,
        CLAHE_grid_size: int=1,
        CLAHE_clipLimit: int=1,
        CLAHE_normalize: bool=True,
    ):
        """
        Augment the FOV images by mixing the FOV with the ROI images
         and optionally applying CLAHE.

        Args:
            ims (list of np.ndarray):
                A list of FOV images.
            spatialFootprints (list of scipy.sparse.csr_matrix, optional):
                A list of spatial footprints for each ROI.
                If None, then no mixing will be performed.
            roi_FOV_mixing_factor (float, optional):
                The factor by which to mix the ROI images into the FOV images.
                If 0, then no mixing will be performed.
            use_CLAHE (bool, optional):
                Whether to apply CLAHE to the images.
            CLAHE_grid_size (int, optional):
                The grid size for CLAHE.
                See alignment.clahe for more details.
            CLAHE_clipLimit (int, optional):
                The clip limit for CLAHE.
                See alignment.clahe for more details.
            CLAHE_normalize (bool, optional):
                Whether to normalize the CLAHE output.
                See alignment.clahe for more details.
        """
        ## Warn if roi_FOV_mixing_factor = 0 but spatialFootprints is not None
        if (roi_FOV_mixing_factor == 0) and (spatialFootprints is not None):
            warnings.warn("roi_FOV_mixing_factor = 0 but spatialFootprints is not None. The ROI images will not be used.")
        ## Warn if roi_FOV_mixing_factor != 0 but spatialFootprints is None
        if (roi_FOV_mixing_factor != 0) and (spatialFootprints is None):
            warnings.warn("roi_FOV_mixing_factor != 0 but spatialFootprints is None. No mixing will be performed.")
        
        h,w = ims[0].shape
        sf = spatialFootprints
        FOV_images = [f + np.array(roi_FOV_mixing_factor*(sf.multiply(1/sf.max(1).A)).sum(0).reshape(h, w)) for f, sf in zip(ims, sf)] if spatialFootprints is not None else ims
        FOV_images = [clahe(im, grid_size=CLAHE_grid_size, clipLimit=CLAHE_clipLimit, normalize=CLAHE_normalize) for im in FOV_images] if use_CLAHE else FOV_images

        self.FOV_images_augmented = FOV_images
        return self.FOV_images_augmented

    def fit_geometric(
        self,
        template: typing.Union[int, np.ndarray],
        ims_moving: typing.List[np.ndarray],
        template_method: str = 'sequential',
        mode_transform: str = 'affine',
        gaussFiltSize: int = 11,
        mask_borders: typing.Tuple[int, int, int, int] = (0, 0, 0, 0),
        n_iter: int = 1000,
        termination_eps: float = 1e-9,
        auto_fix_gaussFilt_step: int = 10,
    ) -> np.ndarray:
        """
        Perform geometric registration of ims_moving to a template.
        Currently relies on cv2.findTransformECC.
        RH 2023

        Args:
            template (int or np.ndarray):
                If 'template_method' == 'image', this is either a 2D np.ndarray
                 image, an integer index of the image to use as the template, or
                 a float between 0 and 1 representing the fractional index of the
                 image to use as the template.
                If 'template_method' == 'sequential', then template is the
                 integer index or fractional index of the image to use as the
                 template.
            ims_moving (list of np.ndarray):
                A list of images to be aligned.
            template_method (str, optional):
                The method to use for template selection.
                'image': use the image specified by 'template'.
                'sequential': register each image to the previous or next image
                    (will be next for images before the template and previous for
                    images after the template)
            mode_transform (str, optional):
                The mode of geometric transformation.
                Can be 'translation', 'euclidean', 'affine', or 'homography'.
                See cv2.findTransformECC for more details.
            gaussFiltSize (int, optional):
                The size of the Gaussian filter, default is 11.
            mask_borders (tuple of int, int, optional):
                The border mask for the image, default is (0, 0, 0, 0).
                (top, bottom, left, right)
            n_iter (int, optional):
                The number of iterations for cv2.findTransformECC, default is 1000.
            termination_eps (float, optional):
                The termination criteria for cv2.findTransformECC, default is 1e-9.
            auto_fix_gaussFilt_step (bool, int):
                Whether to automatically fix convergence issues by increasing the
                 gaussFiltSize.
                If False, then no automatic fixing is performed.
                If True, then the gaussFiltSize is increased by 2 until convergence.
                If int, then the gaussFiltSize is increased by this amount until
                 convergence.

        Returns:
            remapIdx_geo (np.ndarray): 
                An array of shape (N, H, W, 2) representing the remap field for N images.
        """
        ## Imports
        super().__init__()
        
        # Check if ims_moving is a non-empty list
        assert len(ims_moving) > 0, "ims_moving must be a non-empty list of images."
        # Check if all images in ims_moving have the same shape
        shape = ims_moving[0].shape
        for im in ims_moving:
            assert im.shape == shape, "All images in ims_moving must have the same shape."
        # Check if template_method is valid
        valid_template_methods = {'sequential', 'image'}
        assert template_method in valid_template_methods, f"template_method must be one of {valid_template_methods}"
        # Check if mode_transform is valid
        valid_mode_transforms = {'translation', 'euclidean', 'affine', 'homography'}
        assert mode_transform in valid_mode_transforms, f"mode_transform must be one of {valid_mode_transforms}"
        # Check if gaussFiltSize is a number (float or int)
        assert isinstance(gaussFiltSize, (float, int)), "gaussFiltSize must be a number."
        # Convert gaussFiltSize to an odd integer
        gaussFiltSize = int(np.round(gaussFiltSize))

        H, W = ims_moving[0].shape
        self._HW = (H,W) if self._HW is None else self._HW

        ims_moving, template = self._fix_input_images(ims_moving=ims_moving, template=template, template_method=template_method)

        self.mask_geo = helpers.mask_image_border(
            im=np.ones((H, W), dtype=np.uint8),
            border_outer=mask_borders,
            mask_value=0,
        )

        print(f'Finding geometric registration warps with mode: {mode_transform}, template_method: {template_method}, mask_borders: {mask_borders is not None}') if self._verbose else None
        warp_matrices_raw = []
        for ii, im_moving in tqdm(enumerate(ims_moving), desc='Finding geometric registration warps', total=len(ims_moving), disable=not self._verbose):
            if template_method == 'sequential':
                ## warp images before template forward (t1->t2->t3->t4)
                if ii < template:
                    im_template = ims_moving[ii+1]
                ## warp template to itself
                elif ii == template:
                    im_template = ims_moving[ii]
                ## warp images after template backward (t4->t3->t2->t1)
                elif ii > template:
                    im_template = ims_moving[ii-1]
            elif template_method == 'image':
                im_template = template
            
            def _safe_find_geometric_transformation(gaussFiltSize):
                try:
                    warp_matrix = helpers.find_geometric_transformation(
                        im_template=im_template,
                        im_moving=im_moving,
                        warp_mode=mode_transform,
                        n_iter=n_iter,
                        mask=self.mask_geo,
                        termination_eps=termination_eps,
                        gaussFiltSize=gaussFiltSize,
                    )
                except Exception as e:
                    if auto_fix_gaussFilt_step:
                        print(f'Error finding geometric registration warp for image {ii}: {e}') if self._verbose else None
                        print(f'Increasing gaussFiltSize by {auto_fix_gaussFilt_step} to {gaussFiltSize + auto_fix_gaussFilt_step}') if self._verbose else None
                        return _safe_find_geometric_transformation(gaussFiltSize + auto_fix_gaussFilt_step)

                    print(f'Error finding geometric registration warp for image {ii}: {e}')
                    print(f'Defaulting to identity matrix warp.')
                    print(f'Consider doing one of the following:')
                    print(f'  - Make better images to input. You can add the spatialFootprints images to the FOV images to make them better.')
                    print(f'  - Increase the gaussFiltSize parameter. This will make the images blurrier, but may help with registration.')
                    print(f'  - Decrease the termination_eps parameter. This will make the registration less accurate, but may help with registration.')
                    print(f'  - Increase the mask_borders parameter. This will make the images smaller, but may help with registration.')
                    warp_matrix = np.eye(3)[:2,:]
                return warp_matrix
            
            warp_matrix = _safe_find_geometric_transformation(gaussFiltSize=gaussFiltSize)
            warp_matrices_raw.append(warp_matrix)

        
        # compose warp transforms
        print('Composing geometric warp matrices...') if self._verbose else None
        self.warp_matrices = []
        if template_method == 'sequential':
            ## compose warps before template forward (t1->t2->t3->t4)
            for ii in np.arange(0, template):
                warp_composed = self._compose_warps(
                    warp_0=warp_matrices_raw[ii], 
                    warps_to_add=warp_matrices_raw[ii+1:template+1],
                    warpMat_or_remapIdx='warpMat',
                )
                self.warp_matrices.append(warp_composed)
            ## compose template to itself
            self.warp_matrices.append(warp_matrices_raw[template])
            ## compose warps after template backward (t4->t3->t2->t1)
            for ii in np.arange(template+1, len(ims_moving)):
                warp_composed = self._compose_warps(
                    warp_0=warp_matrices_raw[ii], 
                    warps_to_add=warp_matrices_raw[template:ii][::-1],
                    warpMat_or_remapIdx='warpMat',
                )
                self.warp_matrices.append(warp_composed)
        ## no composition when template_method == 'image'
        elif template_method == 'image':
            self.warp_matrices = warp_matrices_raw

        self.warp_matrices = np.stack(self.warp_matrices, axis=0)

        # convert warp matrices to remap indices
        self.remappingIdx_geo = np.stack([helpers.warp_matrix_to_remappingIdx(warp_matrix=warp_matrix, x=W, y=H) for warp_matrix in self.warp_matrices], axis=0)

        return self.remappingIdx_geo
        

    def fit_nonrigid(
        self,
        template: typing.Union[int, np.ndarray],
        ims_moving: typing.List[np.ndarray],
        remappingIdx_init: np.ndarray = None,
        template_method: str = 'sequential',
        mode_transform: str = 'createOptFlow_DeepFlow',
        kwargs_mode_transform: dict = None,
    ) -> np.ndarray:
        """
        Perform geometric registration of ims_moving to a template.
        Currently relies on cv2.findTransformECC.
        RH 2023

        Args:
            template (int or np.ndarray):
                If 'template_method' == 'image', then template is either
                 an image or an integer index or a float fractional index
                 of the image to use as the template.
                If 'template_method' == 'sequential', then template is the
                 integer index of the image to use as the template.
            ims_moving (list of np.ndarray):
                A list of images to be aligned.
            remappingIdx_init (np.ndarray, optional):
                An array of shape (N, H, W, 2) representing any initial
                 remap field to apply to the images in ims_moving.
                The output of this method will be added/composed with remappingIdx_init.
            template_method (str, optional):
                The method to use for template selection.
                'image': use the image specified by 'template'.
                'sequential': register each image to the previous or next image
                    (will be next for images before the template and previous for
                    images after the template)
            mode_transform (str, optional):
                The type of transformation to use for registration.
                Either 'createOptFlow_DeepFlow' or 'calcOpticalFlowFarneback'.
            kwargs_mode_transform (dict, optional):
                Keyword arguments to pass to the mode_transform function.
                See cv2 functions for more details.

        Returns:
            remapIdx_nonrigid (np.ndarray): 
                An array of shape (N, H, W, 2) representing the remap field for N images.
        """
        # Check if ims_moving is a non-empty list
        assert len(ims_moving) > 0, "ims_moving must be a non-empty list of images."
        # Check if all images in ims_moving have the same shape
        shape = ims_moving[0].shape
        for im in ims_moving:
            assert im.shape == shape, "All images in ims_moving must have the same shape."
        # Check if template_method is valid
        valid_template_methods = {'sequential', 'image'}
        assert template_method in valid_template_methods, f"template_method must be one of {valid_template_methods}"
        # Check if mode_transform is valid
        valid_mode_transforms = {'createOptFlow_DeepFlow', 'calcOpticalFlowFarneback'}
        assert mode_transform in valid_mode_transforms, f"mode_transform must be one of {valid_mode_transforms}"

        # Warn if any images have values below 0 or NaN
        found_0 = np.any([np.any(im < 0) for im in ims_moving])
        found_nan = np.any([np.any(np.isnan(im)) for im in ims_moving])
        warnings.warn(f"Found images with values below 0: {found_0}. Found images with NaN values: {found_nan}") if found_0 or found_nan else None

        H, W = ims_moving[0].shape
        self._HW = (H,W) if self._HW is None else self._HW
        x_grid, y_grid = np.meshgrid(np.arange(0., W).astype(np.float32), np.arange(0., H).astype(np.float32))

        ims_moving, template = self._fix_input_images(ims_moving=ims_moving, template=template, template_method=template_method)
        norm_factor = np.nanmax([np.nanmax(im) for im in ims_moving])
        template_norm   = np.array(template * (template > 0) * (1/norm_factor) * 255, dtype=np.uint8) if template_method == 'image' else None
        ims_moving_norm = [np.array(im * (im > 0) * (1/np.nanmax(im)) * 255, dtype=np.uint8) for im in ims_moving]

        print(f'Finding nonrigid registration warps with mode: {mode_transform}, template_method: {template_method}') if self._verbose else None
        remappingIdx_raw = []
        for ii, im_moving in tqdm(enumerate(ims_moving_norm), desc='Finding nonrigid registration warps', total=len(ims_moving_norm), unit='image', disable=not self._verbose):
            if template_method == 'sequential':
                ## warp images before template forward (t1->t2->t3->t4)
                if ii < template:
                    im_template = ims_moving_norm[ii+1]
                ## warp template to itself
                elif ii == template:
                    im_template = ims_moving_norm[ii]
                ## warp images after template backward (t4->t3->t2->t1)
                elif ii > template:
                    im_template = ims_moving_norm[ii-1]
            elif template_method == 'image':
                im_template = template_norm

            if mode_transform == 'calcOpticalFlowFarneback':
                self._kwargs_method = {
                    'pyr_scale': 0.3, 
                    'levels': 3,
                    'winsize': 128, 
                    'iterations': 7,
                    'poly_n': 7, 
                    'poly_sigma': 1.5,
                    'flags': cv2.OPTFLOW_FARNEBACK_GAUSSIAN
                } if kwargs_mode_transform is None else kwargs_mode_transform
                flow_tmp = cv2.calcOpticalFlowFarneback(
                    prev=im_template,
                    next=im_moving, 
                    flow=None, 
                    **self._kwargs_method,
                )
        
            elif mode_transform == 'createOptFlow_DeepFlow':
                flow_tmp = cv2.optflow.createOptFlow_DeepFlow().calc(
                    im_template,
                    im_moving,
                    None
                )

            remappingIdx_raw.append(flow_tmp + np.stack([x_grid, y_grid], axis=-1))

        # compose warp transforms
        print('Composing nonrigid warp matrices...') if self._verbose else None
        
        self.remappingIdx_nonrigid = []
        if template_method == 'sequential':
            ## compose warps before template forward (t1->t2->t3->t4)
            for ii in np.arange(0, template):
                warp_composed = self._compose_warps(
                    warp_0=remappingIdx_raw[ii], 
                    warps_to_add=remappingIdx_raw[ii+1:template+1],
                    warpMat_or_remapIdx='remapIdx',
                )
                self.remappingIdx_nonrigid.append(warp_composed)
            ## compose template to itself
            self.remappingIdx_nonrigid.append(remappingIdx_raw[template])
            ## compose warps after template backward (t4->t3->t2->t1)
            for ii in np.arange(template+1, len(ims_moving)):
                warp_composed = self._compose_warps(
                    warp_0=remappingIdx_raw[ii], 
                    warps_to_add=remappingIdx_raw[template:ii][::-1],
                    warpMat_or_remapIdx='remapIdx',
                    )
                self.remappingIdx_nonrigid.append(warp_composed)
        ## no composition when template_method == 'image'
        elif template_method == 'image':
            self.remappingIdx_nonrigid = remappingIdx_raw

        if remappingIdx_init is not None:
            self.remappingIdx_nonrigid = [self._compose_warps(warp_0=remappingIdx_init[ii], warps_to_add=[warp], warpMat_or_remapIdx='remapIdx') for ii, warp in enumerate(self.remappingIdx_nonrigid)]

        self.remappingIdx_nonrigid = np.stack(self.remappingIdx_nonrigid, axis=0)

        return self.remappingIdx_nonrigid
        
    def transform_images_geometric(self, ims_moving, remappingIdx=None):
        remappingIdx = self.remappingIdx_geo if remappingIdx is None else remappingIdx
        print('Applying geometric registration warps to images...') if self._verbose else None
        self.ims_registered_geo = self._transform_images(ims_moving=ims_moving, remappingIdx=remappingIdx)
        return self.ims_registered_geo
    def transform_images_nonrigid(self, ims_moving, remappingIdx=None):
        remappingIdx = self.remappingIdx_nonrigid if remappingIdx is None else remappingIdx
        print('Applying nonrigid registration warps to images...') if self._verbose else None
        self.ims_registered_nonrigid = self._transform_images(ims_moving=ims_moving, remappingIdx=remappingIdx)
        return self.ims_registered_nonrigid
    
    def _transform_images(self, ims_moving, remappingIdx):
        ims_registered = []
        for ii, (im_moving, remapIdx) in enumerate(zip(ims_moving, remappingIdx)):
            im_registered = helpers.remap_images(
                images=im_moving,
                remappingIdx=remapIdx,
                backend='cv2',
                interpolation_method='linear',
                border_mode='constant',
                border_value=float(im_moving.mean()),
            )
            ims_registered.append(im_registered)
        return ims_registered
    

    def _compose_warps(self, warp_0, warps_to_add, warpMat_or_remapIdx='remapIdx'):
        """
        Compose a series of warps into a single warp.
        RH 2023
        """
        if warpMat_or_remapIdx == 'warpMat':
            fn_compose = helpers.compose_transform_matrices
        elif warpMat_or_remapIdx == 'remapIdx':
            fn_compose = helpers.compose_remappingIdx
        else:
            raise ValueError(f'warpMat_or_remapIdx must be one of ["warpMat", "remapIdx"]')
        
        if len(warps_to_add) == 0:
            return warp_0
        else:
            warp_out = warp_0.copy()
            for warp_to_add in warps_to_add:
                warp_out = fn_compose(warp_out, warp_to_add)
            return warp_out


    def transform_ROIs(
        self, 
        ROIs, 
        remappingIdx=None,
        normalize=True,
    ):
        if remappingIdx is None:
            assert (self.remappingIdx_geo is not None) or (self.remappingIdx_nonrigid is not None), 'If remappingIdx is not provided, then geometric or nonrigid registration must be performed first.'
            remappingIdx = self.remappingIdx_nonrigid if self.remappingIdx_nonrigid is not None else self.remappingIdx_geo

        H, W = remappingIdx[0].shape[:2]

        print('Registering ROIs...') if self._verbose else None
        self.ROIs_aligned = []
        for ii, (remap, rois) in tqdm(enumerate(zip(remappingIdx, ROIs)), total=len(remappingIdx), mininterval=1, disable=not self._verbose, desc='Registering ROIs', position=1):
            rois_aligned = helpers.remap_sparse_images(
                ims_sparse=[roi.reshape((H, W)) for roi in rois],
                remappingIdx=remap,
                method='cubic',
                fill_value=0,
                dtype=np.float32,
                safe=True,
            )
            rois_aligned = scipy.sparse.vstack([roi.reshape(1, -1) for roi in rois_aligned])

            if normalize:
                rois_aligned.data[rois_aligned.data < 0] = 0
                rois_aligned = rois_aligned.multiply(1/rois_aligned.sum(1))
                rois_aligned.data[np.isnan(rois_aligned.data)] = 0
                rois_aligned.eliminate_zeros()
            
            ## remove NaNs from ROIs
            rois_aligned = rois_aligned.tocsr()
            rois_aligned.data[np.isnan(rois_aligned.data)] = 0

            self.ROIs_aligned.append(rois_aligned)

        return self.ROIs_aligned

    def get_ROIsAligned_maxIntensityProjection(self, H=None, W=None):
        """
        Returns the max intensity projection of the ROIs aligned to the template FOV.
        """
        if H is None:
            assert self._HW is not None, 'H and W must be provided if not already set.'
            H, W = self._HW
        return [rois.max(0).toarray().reshape(H, W) for rois in self.ROIs_aligned]
    
    def get_flowFields(self, remappingIdx=None):
        if remappingIdx is None:
            assert (self.remappingIdx_geo is not None) or (self.remappingIdx_nonrigid is not None), 'If remappingIdx is not provided, then geometric or nonrigid registration must be performed first.'
            remappingIdx = self.remappingIdx_nonrigid if self.remappingIdx_nonrigid is not None else self.remappingIdx_geo
        return [helpers.remappingIdx_to_flowField(remap) for remap in remappingIdx]
    
    def _fix_input_images(
        self,
        ims_moving: typing.List[np.ndarray],
        template: typing.Union[int, np.ndarray],
        template_method: str
    ) -> typing.Tuple[int, typing.List[np.ndarray]]:
        ## convert images to float32 and warn if they are not
        print(f'WARNING: ims_moving are not all dtype: np.float32, found {np.unique([im.dtype for im in ims_moving])}, converting...') if any(im.dtype != np.float32 for im in ims_moving) else None
        ims_moving = [im.astype(np.float32) for im in ims_moving]    

        if template_method == 'image':
            if isinstance(template, int):
                assert 0 <= template < len(ims_moving), f'template must be between 0 and {len(ims_moving)-1}, not {template}'
                print(f'WARNING: template image is not dtype: np.float32, found {ims_moving[template].dtype}, converting...') if ims_moving[template].dtype != np.float32 else None
                template = ims_moving[template]
            elif isinstance(template, float):
                assert 0.0 <= template <= 1.0, f'template must be between 0.0 and 1.0, not {template}'
                idx = int(len(ims_moving) * template)
                print(f'Converting float fractional index to integer index: {template} -> {idx}')
                template = ims_moving[idx] # take the image at the specified fractional index
            elif isinstance(template, np.ndarray):
                assert template.ndim == 2, f'template must be 2D, not {template.ndim}'
            else:
                raise ValueError(f'template must be np.ndarray or int or float between 0.0-1.0, not {type(template)}')
            if template.dtype != np.float32:
                print(f'WARNING: template image is not dtype: np.float32, found {template.dtype}, converting...')
                template = template.astype(np.float32)        

        elif template_method == 'sequential':
            assert isinstance(template, (int, float)), f'template must be int or float between 0.0-1.0, not {type(template)}'
            if isinstance(template, float):
                assert 0.0 <= template <= 1.0, f'template must be between 0.0 and 1.0, not {template}'
                idx = int(len(ims_moving) * template)
                print(f'Converting float fractional index to integer index: {template} -> {idx}')
                template = idx
            assert 0 <= template < len(ims_moving), f'template must be between 0 and {len(ims_moving)-1}, not {template}'

        return ims_moving, template


class PhaseCorrelation_registration:
    """
    Class for performing rigid transformation using phase correlation.
    RH 2022
    """
    def __init__(
        self,
    ):
        self.mask = None

    def set_spectral_mask(
        self,
        freq_highPass=0.01,
        freq_lowPass=0.3,
        im_shape=(512, 512),
    ):
        self.mask = make_spectral_mask(
            freq_highPass=freq_highPass,
            freq_lowPass=freq_lowPass,
            im_shape=im_shape,
        )

    def register(
        self, 
        template, 
        ims_moving,
        template_method='sequential',
    ):
        """
        Register set of images using phase correlation.
        RH 2022
        
        Args:
            template (np.ndarray or int):
                Template image
            ims_moving (np.ndarray):
                Images to align to the template.
            template_method (str):
                The method used to register the images.
                Either 'image' or 'sequential'.
                If 'image':      template must be a single image.
                If 'sequential': template must be an integer corresponding 
                 to the index of the image to set as 'zero' offset.
        
        Returns:
            template (np.ndarray):
                Registered images
            shifts (np.ndarray):
                Pixel shift values (y, x).

        Attributes set:
            self.ims_registered (np.ndarray):
                Registered images
            self.shifts (np.ndarray):
                Pixel shift values (y, x).
            self.ccs (np.ndarray):
                Phase correlation coefficient images.
            self.ims_template_filt (np.ndarray):
                Template images filtered by the spectral mask.
            self.ims_moving_filt (np.ndarray):
                Moving images filtered by the spectral mask.
        """
        self.ccs, self.ims_template_filt, self.ims_moving_filt, self.shifts, self.ims_registered = [], [], [], [], []
        shift_old = np.array([0,0])
        for ii, im_moving in enumerate(ims_moving):
            if template_method == 'sequential':
                im_template = ims_moving[ii-1] if ii > 0 else ims_moving[ii]
            elif template_method == 'image':
                im_template = template

            cc, im_template_filt, im_moving_filt = phase_correlation(
                im_template,
                im_moving,
                mask_fft=self.mask,
                return_filtered_images=True,
            )
            shift = convert_phaseCorrelationImage_to_shifts(cc) + shift_old

            self.ccs.append(cc)
            self.ims_template_filt.append(im_template_filt)
            self.ims_moving_filt.append(im_moving_filt)
            self.shifts.append(shift)

            shift_old = shift.copy() if template_method == 'sequential' else shift_old
        
        self.shifts = np.stack(self.shifts, axis=0)
        self.shifts = self.shifts - self.shifts[template,:] if template_method == 'sequential' else self.shifts

        if template_method == 'sequential':
            self.ims_registered = [shift_along_axis(shift_along_axis(im_moving, shift[0] - self.shifts[template,0], fill_val=0, axis=0), shift[1] - self.shifts[template,1], fill_val=0, axis=1) for ii, (im_moving, shift) in enumerate(zip(ims_moving, self.shifts))]
        elif template_method == 'image':
            self.ims_registered = [shift_along_axis(shift_along_axis(im_moving, shift[0], fill_val=0, axis=0), shift[1], fill_val=0, axis=1) for ii, (im_moving, shift) in enumerate(zip(ims_moving, self.shifts))]

        return self.ims_registered, self.shifts


def phase_correlation(im_template, im_moving, mask_fft=None, return_filtered_images=False):
    """
    From BNPM
    Perform phase correlation on two images.
    RH 2022
    
    Args:
        im_template (np.ndarray):
            Template image
        im_moving (np.ndarray):
            Moving image
        mask_fft (np.ndarray):
            Mask for the FFT.
            If None, no mask is used.
    
    Returns:
        cc (np.ndarray):
            Phase correlation coefficient.
    """
    
    if mask_fft is None:
        mask_fft = np.ones(im_template.shape)
    else:
        mask_fft = np.fft.fftshift(mask_fft/mask_fft.sum())

    fft_template = np.fft.fft2(im_template) * mask_fft
    fft_moving   = np.fft.fft2(im_moving) * mask_fft
    R = np.conj(fft_template) * fft_moving
    R[mask_fft != 0] /= np.abs(R)[mask_fft != 0]
    cc = np.fft.fftshift(np.fft.ifft2(R)).real
    if return_filtered_images == False:
        return cc
    else:
        return cc, np.abs(np.fft.ifft2(fft_template)), np.abs(np.fft.ifft2(fft_moving))


def convert_phaseCorrelationImage_to_shifts(cc_im):
    """
    Convert phase correlation image to pixel shift values.
    RH 2022

    Args:
        cc_im (np.ndarray):
            Phase correlation image.
            Middle of image is zero-shift.

    Returns:
        shifts (np.ndarray):
            Pixel shift values (y, x).
    """
    height, width = cc_im.shape
    shift_y_raw, shift_x_raw = np.unravel_index(cc_im.argmax(), cc_im.shape)
    return int(np.floor(height/2) - shift_y_raw) , int(np.ceil(width/2) - shift_x_raw)


def helper_shift(X, shift, fill_val=0):
    X_shift = np.empty_like(X, dtype=X.dtype)
    if shift>0:
        X_shift[:shift] = fill_val
        X_shift[shift:] = X[:-shift]
    elif shift<0:
        X_shift[shift:] = fill_val
        X_shift[:shift] = X[-shift:]
    else:
        X_shift[:] = X
    return X_shift
def shift_along_axis(X, shift, fill_val=0, axis=0):
    return np.apply_along_axis(helper_shift, axis, np.array(X, dtype=X.dtype), shift, fill_val)


def make_spectral_mask(freq_highPass=0.01, freq_lowPass=0.3, im_shape=(512, 512)):
    height, width = im_shape[0], im_shape[1]
    
    idx_highPass = (int(np.ceil(height * freq_highPass / 2)), int(np.ceil(width * freq_highPass / 2)))
    idx_lowPass = (int(np.floor(height * freq_lowPass / 2)), int(np.floor(width * freq_lowPass / 2)))

    if freq_lowPass < 1:
        mask = np.ones((height, width))
        mask[idx_lowPass[0]:-idx_lowPass[0]+1,:] = 0
        mask[:,idx_lowPass[1]:-idx_lowPass[1]+1] = 0
    else:
        mask = np.ones((height, width))

    mask_high = np.fft.fftshift(mask)

    if (idx_highPass[0] > 0) and (idx_highPass[1] > 0):
        mask = np.zeros((height, width))
        mask[idx_highPass[0]:-idx_highPass[0],:] = 1
        mask[:,idx_highPass[1]:-idx_highPass[1]] = 1
    else:
        mask = np.ones((height, width))

    mask_low = np.fft.fftshift(mask)

    mask_out = mask_high * mask_low
    
    return mask_out


def clahe(im, grid_size=50, clipLimit=0, normalize=True):
    """
    Perform Contrast Limited Adaptive Histogram Equalization (CLAHE)
     on an image.
    RH 2022

    Args:
        im (np.ndarray):
            Input image
        grid_size (int):
            Grid size.
            See cv2.createCLAHE for more info.
        clipLimit (int):
            Clip limit.
            See cv2.createCLAHE for more info.
        normalize (bool):
            Whether to normalize the output image.
        
    Returns:
        im_out (np.ndarray):
            Output image
    """
    im_tu = (im / im.max())*(2**8) if normalize else im
    im_tu = (im_tu/10).astype(np.uint8)
    clahe = cv2.createCLAHE(clipLimit=clipLimit, tileGridSize=(grid_size, grid_size))
    im_c = clahe.apply(im_tu.astype(np.uint16))
    return im_c



import numpy as np
import torch
import joblib
from tqdm.notebook import tqdm
from sklearn.linear_model import LogisticRegression
import matplotlib.pyplot as plt
import seaborn as sns
from . import pipeline, evaluate
from .. import helpers
from copy import deepcopy
from .. import visualization as cviz

class CrossValidation():
    """
    Class for loading and splitting data for the purposes of Cross Validation
    JZ 2022
    """
    def __init__(self, splitter):
        """
        Construct the CrossValidation object
        Args:
            splitter (sklearn object inheriting from BaseShuffleSplit):
             Splitter object to use to split data for validation
        """
        self.splitter = splitter
        self.split_inx = None
    
    def split_to_inx(self, *args, **kwargs):
        """
        Generate list of split indices to be used in validation splits
        Args:
            args: Positional arguments to feed into self.splitter.split
            kwargs: Keword arguments to feed into self.splitter.split
        """
        self.split_inx = list(self.splitter.split(*args, **kwargs))
    
    def inx_to_data(self, x, y):
        """
        Generate lists of x/y, training/validation data for validation loop
        Args:
            x (np.array): Predictive features (examples along axis=0)
            y (np.array): Response labels (examples along axis=0)
        """
        self.split_x_tr = [x[_[0]] for _ in self.split_inx]
        self.split_x_val = [x[_[1]] for _ in self.split_inx]
        self.split_y_tr = [y[_[0]] for _ in self.split_inx]
        self.split_y_val = [y[_[1]] for _ in self.split_inx]
    
#     def get_inx_data(self, inx):
#         return self.split_x_tr[inx], self.split_x_val[inx], self.split_y_tr[inx], self.split_y_val[inx]
    
    def get_inx_data(self):
        """
        Return a list of zipped validation splits for x/y, training/validation data
        for validation iteration
        
        JZ 2022
        """
        return list(zip(self.split_x_tr, self.split_y_tr, self.split_x_val, self.split_y_val))
    

def split_loop_c(cv, preproc_init, classifier_kwargs, preproc_refit=True, c_lst=[1, 0.1, 0.01], class_weights=None):
    """
    Return a list of zipped validation splits for x/y, training/validation data
    for validation iteration
    
    JZ 2022
    
    Args:
        cv (CrossValidation object):
         Object for generating validation iterator
        preproc_init (sklearn Transform object):
         Initial preprocessor Pipeline pre-fitting
        classifier_kwargs (dict):
         Keword arguments to be used in constructing the LogisticRegression classifier
        preproc_refit (bool):
         Whether or not to refit the preprocessing pipeline on every iteration of the validation loop
        c_lst (list):
         Whether or not to refit the preprocessing pipeline on every iteration of the validation loop
    
    Returns:
        Dictionary of confustion matrices & accuracies from the validation loop
    """
    results_dct = {}
    for ic, c in enumerate(c_lst):
        cm_train, cm_val, acc_train, acc_val = [], [], [], []

        for tmp_trainp_X, tmp_trainp_y, tmp_val_X, tmp_val_y in tqdm(cv.get_inx_data()):
            
            classify_init = LogisticRegression(**classifier_kwargs, C=c)
            pipe = pipeline.fit_pipe(tmp_trainp_X, tmp_trainp_y, preproc_init, classify_init, preproc_refit=preproc_refit, class_weights=class_weights)
            
            evaluator = evaluate.Evaluation(pipe.pipeline)
            cm_train.append(evaluator.confusion_matrix(tmp_trainp_X, tmp_trainp_y))
            cm_val.append(evaluator.confusion_matrix(tmp_val_X, tmp_val_y))

            acc_train.append(evaluator.score_classifier_logreg(tmp_trainp_X, tmp_trainp_y, class_weights=class_weights))
            acc_val.append(evaluator.score_classifier_logreg(tmp_val_X, tmp_val_y, class_weights=class_weights))
        
        results_dct[f'cm_{c}'] = (np.round(np.mean(cm_train,axis=0),2), np.round(np.mean(cm_val,axis=0),2), np.round(np.mean(acc_train, axis=0),2), np.round(np.mean(acc_val, axis=0),2))
    
    return results_dct
    

def view_cv_dict(cv_dct):
    """
    Plot all confusion matrices stored in the cv_dct
    
    JZ 2022
    
    Args:
        cv_dct (dict):
         Dictionary of tuples associated with c-value classifier performance metrics
         consisting of averaged (train CM, val CM, train acc, val acc)
    """
    def confusion_matrix(cm, **params):
        default_params = dict(
            annot=True,
            annot_kws={"size": 16},
            vmin=0.,
            vmax=1.,
            cmap=plt.get_cmap('gray')
        )
        for key in params:
            default_params[key] = params[key]
        sns.heatmap(cm, **default_params)

    for k in cv_dct:
        c = k.replace('cm_','')
        fig, ax = plt.subplots(1, 2, figsize=(10,3))
        fig.set_facecolor('white')
        fig.suptitle(f'C: {c}')
        
#         sns.heatmap(cv_dct[k][0], annot=True, annot_kws={"size": 16}, vmax=1., cmap=plt.get_cmap('gray'), ax=ax[0])
#         sns.heatmap(cv_dct[k][1], annot=True, annot_kws={"size": 16}, vmax=1., cmap=plt.get_cmap('gray'), ax=ax[1])
        
        confusion_matrix(np.round(cv_dct[k][0],4), ax=ax[0])
        confusion_matrix(np.round(cv_dct[k][1],4), ax=ax[1])
        
        ax[0].set_title(f'Train — Acc: {cv_dct[k][2]}')
        ax[1].set_title(f'Val — Acc: {cv_dct[k][3]}')
